import PQueue from 'p-queue';
import { CanAppend, Identity, Log } from '@dao-xyz/ipfs-log';
import { IPFS } from 'ipfs-core-types';
import { Entry } from '@dao-xyz/ipfs-log';
import { EntryWithRefs } from './entry-with-refs';
interface Store<T> {
    _oplog: Log<T>;
    _ipfs: IPFS;
    identity: Identity;
    canAppend?: CanAppend<T>;
}
export declare class Replicator<T> {
    _store: Store<T>;
    _concurrency: number;
    _q: PQueue;
    _logs: Log<T>[];
    _fetching: any;
    _fetched: any;
    onReplicationComplete?: (logs: Log<any>[]) => void;
    onReplicationQueued?: (entry: Entry<T>) => void;
    onReplicationProgress?: (entry: Entry<T>) => void;
    constructor(store: Store<any>, concurrency?: number);
    /**
     * Returns the number of replication tasks running currently
     * @return {[Integer]} [Number of replication tasks running]
     */
    get tasksRunning(): number;
    /**
     * Returns the number of replication tasks currently queued
     * @return {[Integer]} [Number of replication tasks queued]
     */
    get tasksQueued(): number;
    /**
     * Returns the hashes currently queued or being processed
     * @return {[Array]} [Strings of hashes of entries currently queued or being processed]
     */
    get unfinished(): string[];
    load(entries: (Entry<T> | EntryWithRefs<T> | string)[]): Promise<void>;
    _addToQueue(entries: (Entry<T> | EntryWithRefs<T> | string)[]): Promise<void>;
    stop(): Promise<void>;
    start(): Promise<void>;
    _replicateLog(entry: EntryWithRefs<T> | Entry<T> | string): Promise<Log<T>>;
}
export {};
